﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using System.Windows.Forms;

namespace BryanKellyCapstone2017
{   /*
    WRITTEN BY BRYAN KELLY AS PART OF THE 2017 CAPSTONE SPECIFICATIONS GIVEN BY DR. DAVID PANKRATZ

    NUMBER MEANINGS
    0 = EMPTY SPACE 
    1 = WALL 
    2 = ROBOT
    3 = TARGET BLOCK
    4 = DESTINATION/GOAL LOCATION

    COLOR ASSOCIATIONS
    WHITE = EMPTY SPACE
    BLACK = WALL
    BLUE = ROBOT
    RED = TARGET BLOCK
    GREEN = DESTINATION/GOAL BLOCK
     */
    public partial class Main : Form
    {
        //ai test variables these should proabably be removed
        private int goalRow = -1;
        private int goalColumn = -1;

        //other cs files
        private FloorMap floorMap;
        private AITestFunctions testAI;

        //graphics stuff
        private Bitmap imgFloor;
        private Graphics imgFloorGraphics;

        //bool
        private bool isFileOpen = false;
        private bool manualMovementEnabled = false;

        //constants
        private const int NUMBRUSHES = 5;
        private const int EMPTY = 0;
        private const int WALL = 1;
        private const int ROBOT = 2;
        private const int TARGET = 3;
        private const int GOAL = 4;

        //global variables
        private int totalRows;
        private int totalColumns;
        private int robotRow;
        private int robotColumn;
        private int moveCount = 0;

        private string fileName = "";

        //Brushes
        private Brush brushEmpty = new SolidBrush(Color.WhiteSmoke);
        private Brush brushWall = new SolidBrush(Color.Black);
        private Brush brushRobot = new SolidBrush(Color.Cyan);
        private Brush brushTarget = new SolidBrush(Color.Maroon);
        private Brush brushGoal = new SolidBrush(Color.Green);
        private Brush[] currentBrush;
        //additional Brush for Grid markers
        private Brush brushGrid = new SolidBrush(Color.LightGray);

        //program Main
        public Main()
        {
            InitializeComponent();
            floorMap = new FloorMap();
            testAI = new AITestFunctions();

            //setup brushes
            currentBrush = new Brush[5];//the number of options for types of squares
            currentBrush[EMPTY] = brushEmpty;
            currentBrush[WALL] = brushWall;
            currentBrush[ROBOT] = brushRobot;
            currentBrush[TARGET] = brushTarget;
            currentBrush[GOAL] = brushGoal;

            //setup graphics just to avoid an error
            imgFloor = new Bitmap(pnlFloorPlan.Width, pnlFloorPlan.Height);
            imgFloorGraphics = Graphics.FromImage(imgFloor);
            pnlFloorPlan.Invalidate();

            //event handlers
            this.floorMap.OpenDocument += new System.EventHandler(this.updateFloorMapOnOpen);
            this.floorMap.UpdateCaption += new System.EventHandler(this.updateStatsLables);
            this.floorMap.targetReachedGoal += new System.EventHandler(this.updatelblNumTargets);
            this.floorMap.allTargetsReachedGoal += new System.EventHandler(this.allTargetsAtGoal);
            this.floorMap.multipleRobotsDetectedInFile += new System.EventHandler(this.displayMultipleRobotMessage);
            //move Robot events
            this.floorMap.moveRobot += new System.EventHandler(this.drawMoveRobot);
            //move Target events
            this.floorMap.moveTarget += new System.EventHandler(this.drawMoveTarget);
            //redraw Goal after robot moves off of spot
            this.floorMap.redrawGoal += new System.EventHandler(this.redrawGoalLocation);

            //some test AI events
            this.testAI.errorTryingToMoveRobot += new System.EventHandler(this.errorTryingToMoveRobot);
            this.testAI.robotReachedDestination += new System.EventHandler(this.robotReachedDestination);

        }

        //move Robot paint Events
        private void drawMoveRobot(object sender, EventArgs e)
        {   /*will fill old robot tile with empty space and will draw robot color on new space*/
            moveCount++;
            lblMoveCount.Text = "Total Moves: " + moveCount;
            drawOverSpace(EMPTY, robotRow, robotColumn);

            updateLogWithMove("Robot", floorMap.RobotRow, floorMap.RobotColumn);

            //update robot row and column
            robotRow = floorMap.RobotRow;
            robotColumn = floorMap.RobotColumn;

            drawOverSpace(ROBOT, robotRow, robotColumn);
            updatelblRobotLocation(robotRow, robotColumn);
        }
        private void updateLogWithMove(string movedObject, int newRow, int newColumn)
        {

            rtbLog.Text += moveCount + ": " + movedObject + " moved to " + newColumn + ", " + newRow + "\n";
        }
        private void allTargetsAtGoal(object sender, EventArgs e)
        {
            MessageBox.Show("All targets have reached a goal. \nCompleted in " + moveCount + " moves");
        }
        private void updatelblRobotLocation(int row, int column)
        {
            if (row >= 0 && column >= 0)
                lblRobotLocation.Text = "Robot Location: " + column + "," + row;
            else
                lblRobotLocation.Text = "Robot Location:";
        }
        private void drawOverSpace(int brushType, int row, int column)
        {   //this function will draw a color on a given space
            //want to avoid drawing over the gridlines
            int x = calculateCoordinate(column) + 1;
            int y = calculateCoordinate(row) + 1;
            int tileSize = calculateTileSize();
            tileSize = tileSize - 1; //want to allow for one pixel of space on each side of square
            Rectangle tempSquare = new Rectangle(x, y, tileSize, tileSize);
            imgFloorGraphics.FillRectangle(currentBrush[brushType], x, y, tileSize, tileSize);
            pnlFloorPlan.Invalidate(tempSquare);
        }

        //move Target paint Events
        private void drawMoveTarget(object sender, EventArgs e)
        {
            drawOverSpace(TARGET, floorMap.TargetNewRow, floorMap.TargetNewColumn);
            updateLogWithMove("Target", floorMap.TargetNewRow, floorMap.TargetNewColumn);
        }
        private void redrawGoalLocation(object sender, EventArgs e)
        {
            drawOverSpace(GOAL, floorMap.GoalRow, floorMap.GoalColumn);
        }

        //file open events
        private void fileOpenClick(object sender, EventArgs e)
        {
            resetRobotData();
            floorMap.Open();
        }
        private void resetRobotData()
        {
            moveCount = 0;
            lblMoveCount.Text = "Total Moves: " + moveCount;
            rtbLog.Text = "";
            robotRow = -1;
            robotColumn = -1;
            updatelblRobotLocation(robotRow, robotColumn);
        }
        private void updateFloorMapOnOpen(object sender, EventArgs e)
        {   //updates the floor map if the file could be opened
            //MessageBox.Show("updateFloorMapOpen reached");
            isFileOpen = true;
            fileName = floorMap.GetFileName;
            drawEntireFloorMap();

        }
        private int calculateTileSize()
        {
            //set rows and columns
            totalRows = floorMap.NumRows;
            totalColumns = floorMap.NumColumns;

            //setSize of columns and rows
            int tileHeight = pnlFloorPlan.Height / totalRows;
            int tileWidth = pnlFloorPlan.Width / totalColumns;

            //decides which is larger rows or columns to make squares square
            if (tileWidth > tileHeight)
                return tileHeight;
            else
                return tileWidth;
        }
        private int calculateCoordinate(int input)
        {
            //calculates an X or Y coorinate based on input of a column or row
            //this could be more efficient by storing the tileSize as a global varialbe. 
            //But writing it this way may make it easier to avoid errors
            return input * calculateTileSize();
        }
        private int calculateCoordinate(int input, int tileSize)
        {   //overloaded function just to be slightly more efficient while hopefully maintaining readability
            return input * tileSize;
        }
        private void drawEntireFloorMap()
        {   /*This function will draw the floor map from the file that was read*/
            //reset panel image from any previous file
            imgFloor = new Bitmap(pnlFloorPlan.Width, pnlFloorPlan.Height);
            imgFloorGraphics = Graphics.FromImage(imgFloor);

            int tileSize = calculateTileSize();

            //loop to draw the board
            for (int row = 0; row < totalRows; row++)
            {
                for (int column = 0; column < totalColumns; column++)
                {
                    int x = calculateCoordinate(column, tileSize);
                    int y = calculateCoordinate(row, tileSize);
                    int brushType = WALL; //default to a wall space
                    string tileType = "";

                    //check to see if the colum exists if the text file wasn't complete for that column
                    if (floorMap.isValidLocation(row, column))
                        tileType = floorMap.FloorLayout[row][column];
                    else
                        tileType = "1"; //WALL if location doesn't exist

                    //need to check if the last spot is a null
                    if (tileType != null && tileType != "")
                        brushType = Convert.ToInt32(tileType);
                    if (brushType >= NUMBRUSHES || brushType < 0)
                        brushType = WALL;
                    if (brushType == ROBOT)
                    {
                        robotRow = row;
                        robotColumn = column;
                        updatelblRobotLocation(robotRow, robotColumn);
                    }
                    if (brushType == GOAL)
                    {   ///Delete this if statement once done!!!!!!!!!!!!!!!!!!!!!!
                        goalRow = row;
                        goalColumn = column;
                    }
                    imgFloorGraphics.FillRectangle(currentBrush[brushType], x, y, tileSize, tileSize);

                }
            }
            drawGridLines();
            pnlFloorPlan.Invalidate();
        }
        private void displayMultipleRobotMessage(object sender, EventArgs e)
        {
            MessageBox.Show("Multiple robots detected in file. Only using first robot found in file. Other robots will be treated as walls.");
        }
        private void updateStatsLables(object sender, EventArgs e)
        {   /* updates the floor size label on the side of the screen */
            lblFloorSize.Text = "Floor Size: " + floorMap.NumColumns + " columns X " + floorMap.NumRows + " rows";
            lblNumTargets.Text = "Targets Remaining:" + floorMap.NumTargets;
        }
        private void updatelblNumTargets(object sender, EventArgs e)
        {
            rtbLog.Text += "A target reached the goal\n" + floorMap.NumTargets + " target blocks remaining.\n";
            lblNumTargets.Text = "Targets Remaining:" + floorMap.NumTargets;
        }

        private void moveRobotNorth(object sender, EventArgs e)
        {   //triggers moveRobotNorthEvent
            if (isFileOpen)
                floorMap.moveRobotNorth();
        }

        private void moveRobotSouth(object sender, EventArgs e)
        {
            if (isFileOpen)
                floorMap.moveRobotSouth();
        }

        private void moveRobotEast(object sender, EventArgs e)
        {
            if (isFileOpen)
                floorMap.moveRobotEast();
        }

        private void moveRobotWest(object sender, EventArgs e)
        {
            if (isFileOpen)
                floorMap.moveRobotWest();
        }

        private void fileReset(object sender, EventArgs e)
        {
            resetRobotData();
            floorMap.Open(fileName);
        }

        private void rtbLogAutoScroll(object sender, EventArgs e)
        {   /*function written by Omar Mahili on stackoverflow.com 
            Link: http://stackoverflow.com/questions/9416608/rich-text-box-scroll-to-the-bottom-when-new-data-is-written-to-it */

            // set the current caret position to the end
            rtbLog.SelectionStart = rtbLog.Text.Length;
            // scroll it automatically
            rtbLog.ScrollToCaret();

        }

        private void drawGridLines()
        { /* this draws gridlines to make the floor map easier to distinguish tiles */
            int tileSize = calculateTileSize();

            for (int row = 0; row <= totalRows; row++)
            {   //draws horizontal gridlines
                int x = 0;
                int y = calculateCoordinate(row, tileSize);
                imgFloorGraphics.FillRectangle(brushGrid, x, y, totalColumns * tileSize, 1);
            }
            for (int column = 0; column <= totalColumns; column++)
            {   //draws vertical gridlines
                int x = calculateCoordinate(column, tileSize);
                int y = 0;
                imgFloorGraphics.FillRectangle(brushGrid, x, y, 1, totalRows * tileSize);
            }
        }//end drawGrid

        private void pnlFloorPlanPaint(object sender, PaintEventArgs e)
        {   /*built from Connect Four Program from CS350*/
            Graphics g = pnlFloorPlan.CreateGraphics();

            g.DrawImage(imgFloor, 0, 0);
        }//end pnlFloorPlanPaint

        private void testMoveRobotToGoal(object sender, EventArgs e)
        {
            BackgroundWorker bw = new BackgroundWorker();
            bw.WorkerReportsProgress = true;

            bw.DoWork += new DoWorkEventHandler(
        delegate (object o, DoWorkEventArgs args)
        {
            testAI.depthFirstSearchRobot(floorMap, goalRow, goalColumn);
        });

            bw.RunWorkerAsync();
        }
        private void errorTryingToMoveRobot(object sender, EventArgs e)
        {
            MessageBox.Show("Error trying to move robot. Stopping now");
        }
        private void robotReachedDestination(object sender, EventArgs e)
        {
            MessageBox.Show("Robot reached goal location");
        }

        protected override bool IsInputKey(Keys keyData)
        {   //from "alpha" on stackOverflow.com
            //Link: http://stackoverflow.com/questions/1646998/up-down-left-and-right-arrow-keys-do-not-trigger-keydown-event
            switch (keyData)
            {
                case Keys.Right:
                case Keys.Left:
                case Keys.Up:
                case Keys.Down:
                    return true;
                case Keys.Shift | Keys.Right:
                case Keys.Shift | Keys.Left:
                case Keys.Shift | Keys.Up:
                case Keys.Shift | Keys.Down:
                    return true;
            }
            return base.IsInputKey(keyData);
        }
        protected override void OnKeyDown(KeyEventArgs e)
        {   //from "alpha" on stackOverflow.com edited by "MicroVirus"
            //Link: http://stackoverflow.com/questions/1646998/up-down-left-and-right-arrow-keys-do-not-trigger-keydown-event

            base.OnKeyDown(e);
            if (manualMovementEnabled)
            {
                switch (e.KeyCode)
                {
                    case Keys.A:
                    case Keys.Left: moveRobotWest(this, e); break;
                    case Keys.D:
                    case Keys.Right: moveRobotEast(this, e); break;
                    case Keys.W:
                    case Keys.Up: moveRobotNorth(this, e); break;
                    case Keys.S:
                    case Keys.Down: moveRobotSouth(this, e); break;
                }
            }
        }

        private void toggleManualMovement(object sender, EventArgs e)
        {
            if (manualMovementEnabled)
            {
                manualMovementEnabled = false;
                showHideManualControls(false);
                tsmiEnableManualMovementToggle.Text = "Enable Manual Movement";
            }
            else
            {
                manualMovementEnabled = true;
                showHideManualControls(true);
                tsmiEnableManualMovementToggle.Text = "Disable Manual Movement";
            }
        }
        private void showHideManualControls(bool state)
        {
            btnMoveEast.Enabled = state;
            btnMoveWest.Enabled = state;
            btnMoveNorth.Enabled = state;
            btnMoveSouth.Enabled = state;
            btnMoveEast.Visible = state;
            btnMoveWest.Visible = state;
            btnMoveNorth.Visible = state;
            btnMoveSouth.Visible = state;
        }

    }
}